// Bentigt GLUT!
// Fr Dev-C++ (also Mingw32) siehe http://sites.netscape.net/ptrpck/glut.htm
// mit GLUT linken: Project/Project options/Further object files or linker options:
// -lopengl32 -lglu32 -lglut32

// Korrekturen fr GLUT in Dev-C++:
#define GLUT_NO_LIB_PRAGMA
#define GLUT_NO_WARNING_DISABLE

#include <gl/glut.h>
#include <math.h>
#include "tga.h"

const double pi = 4.0*atan(1.0);
const int slices = 20; // Scheibchen der Kugel
const int stacks = 20; // Schichten der Kugel
// Koordinaten pro Vertex
double x[stacks+1][slices+1]; // letzter Stack = Sdpol, letztes Slice = erstes Slice
double y[stacks+1][slices+1];
double z[stacks+1][slices+1];
// Normalenvektoren pro Vertex
double n[stacks+1][slices+1];
double m[stacks+1][slices+1];
double o[stacks+1][slices+1];
// Texturkoordinaten pro Vertex
double s[stacks+1][slices+1];
double t[stacks+1][slices+1];

tgaInfo* myTexture;

// schner wren Vertex-Arrays
void buildSphere() // Arrays mit Daten der Einheitskugel laden
{
    int i; // fr MSVC nicht for(int ...)
    for(i=0; i<=stacks; i++) // i bis stacks einschlielich!
    {
        double y_ = 1.0 - i*(2.0/stacks);
        double r_xz = sqrt(1.0-y_*y_); // Radius in xz-Ebene

        int j; // fr MSVC nicht for(int ...)
        for(j=0; j<=slices; j++) // j bis slices einschlielich!
        {
            // Koordinaten berechnen
            double phi = j*(2.0*pi/slices);
            double x_ = r_xz * cos(phi);
            double z_ = r_xz * sin(phi);
            x[i][j] = x_;
            y[i][j] = y_;
            z[i][j] = z_;

            // Normalenvektor berechnen
            double n_ = x_;
            double m_ = y_; // knnte aus der Schleife raus (unbersichtlich)
            double o_ = z_;
            double nLen = sqrt(n_*n_ + m_*m_ + o_*o_); // berflssig, weil hier 1
            n_ /= nLen;
            m_ /= nLen;
            o_ /= nLen;
            n[i][j] = n_;
            m[i][j] = m_;
            o[i][j] = o_;

            // Texturkoordinaten berechnen
            s[i][j] = 2*x_;
            t[i][j] = 2*z_;
        }
    }
}

void renderScene()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Lichtquelle fest
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    GLfloat lightPos[] = {-1.5, 0.8, 3.9, 1.0};
    glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
    glPopMatrix();

    /*
    // langes "Brett", um Wirkung von MIP-Map zu zeigen
    // automatische Generierung der Texturkoordinaten ausschalten!
    glNormal3f(0.0, 1.0, 0.0);
    glBegin(GL_QUADS);
        glTexCoord2f(-2.0, -100.0);
        glVertex3f(-2.0, -2.0, -100.0);

        glTexCoord2f(-2.0, 2.0);
        glVertex3f(-2.0, -2.0, 2.0);

        glTexCoord2f(2.0, 2.0);
        glVertex3f(2.0, -2.0, 2.0);

        glTexCoord2f(2.0, -100.0);
        glVertex3f(2.0, -2.0, -100.0);
    glEnd();
    */

    /*
    // Kugel mit benutzerdefinierter ebener Texturprojektion
    // automatische Generierung der Texturkoordinaten ausschalten!
    int i; // fr MSVC nicht for(int ...)
    for(i=0; i<stacks; i++)
    {
        glBegin(GL_QUADS); // krzer mit GL_QUAD_STRIP
            int j; // fr MSVC nicht for(int ...)
            for(j=0; j<slices; j++)
            {
                // nur zur Demonstration ausgeschrieben, das gehrt
                // selbstverstndlich mit Hilfsfunktionen strukturiert
                glNormal3f(n[i][j], m[i][j], o[i][j]);
                glTexCoord2f(s[i][j], t[i][j]);
                glVertex3f(x[i][j], y[i][j], z[i][j]);

                glNormal3f(n[i][j+1], m[i][j+1], o[i][j+1]);
                glTexCoord2f(s[i][j+1], t[i][j+1]);
                glVertex3f(x[i][j+1], y[i][j+1], z[i][j+1]);

                glNormal3f(n[i+1][j+1], m[i+1][j+1], o[i+1][j+1]);
                glTexCoord2f(s[i+1][j+1], t[i+1][j+1]);
                glVertex3f(x[i+1][j+1], y[i+1][j+1], z[i+1][j+1]);

                glNormal3f(n[i+1][j], m[i+1][j], o[i+1][j]);
                glTexCoord2f(s[i+1][j], t[i+1][j]);
                glVertex3f(x[i+1][j], y[i+1][j], z[i+1][j]);
            }
        glEnd();
    }
    */

    /*
    // Kugel mit automatischer Texturprojektion,
    // Vorher mit glEnable anschalten!
    // ebene Projektion:
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    GLfloat xDir[] = {2.0, 0.0, 0.0, 0.0};
    GLfloat zDir[] = {0.0, 0.0, 2.0, 0.0};
    glTexGenfv(GL_S, GL_OBJECT_PLANE, xDir);
    glTexGenfv(GL_T, GL_OBJECT_PLANE, zDir);

    int i; // fr MSVC nicht for(int ...)
    for(i=0; i<stacks; i++)
    {
        glBegin(GL_QUADS); // krzer mit GL_QUAD_STRIP
            int j; // fr MSVC nicht for(int ...)
            for(j=0; j<slices; j++)
            {
                glNormal3f(n[i][j], m[i][j], o[i][j]);
                glVertex3f(x[i][j], y[i][j], z[i][j]);

                glNormal3f(n[i][j+1], m[i][j+1], o[i][j+1]);
                glVertex3f(x[i][j+1], y[i][j+1], z[i][j+1]);

                glNormal3f(n[i+1][j+1], m[i+1][j+1], o[i+1][j+1]);
                glVertex3f(x[i+1][j+1], y[i+1][j+1], z[i+1][j+1]);

                glNormal3f(n[i+1][j], m[i+1][j], o[i+1][j]);
                glVertex3f(x[i+1][j], y[i+1][j], z[i+1][j]);
            }
        glEnd();
    }
    */

    // sphrische Environment-Map
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glutSolidTeapot(1.0);

    glutSwapBuffers();
}

void changeSize(GLsizei w, GLsizei h)
{
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    gluPerspective(60.0, h==0 ? 1.0 : w/(double)h, 1.0, 103.0);

    glTranslatef(0.0, 0.0, -3.0);
}

void specialFunc(int key, int x, int y)
{
    glMatrixMode(GL_MODELVIEW);

    float oldModelView[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, oldModelView);
    glLoadIdentity();

    switch(key)
    {
    case GLUT_KEY_RIGHT:
        glRotatef(2.0, 0.0, 1.0, 0.0);
        break;
    case GLUT_KEY_LEFT:
        glRotatef(-2.0, 0.0, 1.0, 0.0);
        break;
    case GLUT_KEY_UP:
        glRotatef(-2.0, 1.0, 0.0, 0.0);
        break;
    case GLUT_KEY_DOWN:
        glRotatef(2.0, 1.0, 0.0, 0.0);
        break;
    default:
        ;
    }

    glMultMatrixf(oldModelView);
    glGetFloatv(GL_MODELVIEW_MATRIX, oldModelView);

    glutPostRedisplay();
}

int main(int argc, char* argv[])
{
    myTexture = tgaLoad("textur.tga");
    if(myTexture == 0 || myTexture->status != TGA_OK)
    {
        exit(-1); // nicht gerade benutzerfreundlich
    }

    buildSphere();

    glutInit(&argc, argv);

    glutInitWindowPosition(30, 30);
    glutInitWindowSize(800, 600);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("Texturen");
    glutDisplayFunc(renderScene);
    glutReshapeFunc(changeSize);
    glutSpecialFunc(specialFunc);
    glEnable(GL_DEPTH_TEST);
    // glEnable(GL_CULL_FACE);

    glEnable(GL_LIGHTING);

    GLfloat globAmbCol[] = {0.0, 0.0, 0.0, 1.0};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globAmbCol);

    glEnable(GL_LIGHT0);
    GLfloat diffCol[] = {1.0, 1.0, 0.4, 1.0};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffCol);
    GLfloat specCol[] = {1.0, 1.0, 1.0, 1.0};
    glLightfv(GL_LIGHT0, GL_SPECULAR, specCol);
    GLfloat ambCol[] = {0.2, 0.2, 0.2, 1.0};
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambCol);

    // nur ein Material => knnen wir auch hierhin packen statt in renderScene
    GLfloat matDiffCol[] = {1.0, 1.0, 1.0, 1.0};
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, matDiffCol);
    GLfloat matSpecCol[] = {1.0, 1.0, 1.0, 1.0};
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecCol);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 20.0);
    GLfloat matAmbCol[] = {1.0, 1.0, 1.0, 1.0};
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, matAmbCol);
    GLfloat matEmCol[] = {0.1, 0.0, 0.0, 1.0};
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, matEmCol);

    // Achtung: OpenGL verlangt Texturen mit Kantenlngen = Zweierpotenzen.
    // ggf. +1 fr Rand
    // glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, myTexture->imageData);
    //Parameter: Dimensionalitt, Level (fr MIP-Map), Kanalzahl, Breite, Hhe, Rand, ...)

    // MIP-Map automatisch aufbauen:
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 256, 256, GL_RGB, GL_UNSIGNED_BYTE, myTexture->imageData);

    // was tun, wenn Texel kleiner als Pixel:
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    // was tun, wenn Texel grer als Pixel:
    // glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // fr MIP-Map, Detailgrad stufig:
    // glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    // fr MIP-Map, Detailgrad interpoliert, also trilineare Interpolation:
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

    // nur die Farbe der Textur benutzen, ohne Beleuchtung
    // glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

    // Texturierung anschalten:
    glEnable(GL_TEXTURE_2D);

    // Generierung von Texturkoordinaten anschalten:
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);

    glutMainLoop();

    return 0; // nie erreicht
}
